/*! \file fourierfunc.cpp
 *  \brief Implementation of FourierFunc class for manipulating band-limited
 *   functions on S_n
 */

#include <vector>
#include <iostream>
#include <sstream>

#include "fourierfunc.h"
#include "partition.h"
#include "func.h"
#include "util.h"
#include "fourier.h"

using namespace std;

#define Matrix itpp::mat

FourierFunc::FourierFunc(int _nelements)
 : nelements (_nelements)
{

}

/// \todo do this with band-limiting instead
/// \todo provide a way to construct an uninitialized FourierFunc
FourierFunc FourierFunc::zero(vector<Partition> ind_parts)
{
        FourierFunc f(ind_parts[0].num());

        f.parts = ind_parts;

        f.fill_zero();

        return f;
}

Func FourierFunc::inverse_fourier() const
{
        Func f = Func::zero(nelements);
        /// \todo fix next line for band-limiting
        vector<Partition> parts = Partition::list(nelements);
        double funclen = fact(nelements);

        for (int i = 0; i < funclen; i++) {
                Permutation p = Permutation::from_index(nelements, i);
                for (unsigned int j = 0; j < parts.size(); j++) {
                        Matrix irrep = gz_irrep(parts[j], p);
                        f.set(p, f(p) + (irrep.rows()/funclen)
                                         *elem_mult_sum(irrep, coeffs[j]));
                }
        }

        return f;
}

string FourierFunc::str() const
{
        ostringstream out;

        for (unsigned int i = 0; i < coeffs.size(); i++)
                out << coeffs[i] << endl;

        return out.str();
}

/// \todo build a hash table for this?
Matrix FourierFunc::at(const Partition &part) const
{
        for (unsigned int i = 0; i < coeffs.size(); i++)
                if (parts[i] == part)
                        return coeffs[i];

        // this coefficient is not stored, so it is zero
        int d = Tableau::count(part);
        Matrix c (d, d);
        c.zeros();

        return c;
}

void FourierFunc::set(int i, Matrix m)
{
        coeffs[i] = m;
}

vector<Partition> FourierFunc::ind_parts()
{
        return parts;
}

void FourierFunc::set(const Partition &part, Matrix m)
{
        vector<Partition> parts = Partition::list(nelements);
        for (unsigned int i = 0; i < coeffs.size(); i++)
                if (parts[i] == part) {
                        coeffs[i] = m;
                        return;
                }
        throw "Uh Oh";
}

Func FourierFunc::inverse_fft() const
{
        if (nelements == 1) {
                Func f = Func::zero(1);
                f.set(Permutation::identity(1), coeffs[0](0, 0));
                return f;
        }

        Func f = Func::zero(nelements);
        vector<Partition> parts = Partition::list(nelements);

        for (int i = 0; i < nelements; i++) {
                FourierFunc subft = zero(nelements - 1); 
                for (unsigned int j = 0; j < parts.size(); j++) {
                        Permutation cc = Permutation::cont_cycle(nelements, i,
                                                                 nelements);
                        Matrix m = mult_irrep_left(parts[j], cc.inverse(),
                                                   coeffs[j]);
                        vector<Partition> subparts = parts[j].minus();
                        int offset = 0;
                        for (unsigned int k = 0; k < subparts.size(); k++) {
                                Matrix c = subft.at(subparts[k]);
                                int d = Tableau::count(subparts[k]);
                                double f = (double) Tableau::count(parts[j])
                                           /(nelements*d);
                                // submatrix
                                c += m(offset, offset,
                                       offset + d - 1, offset + d - 1) * f;
                                subft.set(subparts[k], c);
                                offset += d;
                        }
                }
                f += subft.inverse_fft().embed(i);
        }

        return f;
}

void FourierFunc::fill_zero()
{
        coeffs = vector<Matrix>();

        for (unsigned int i = 0; i < parts.size(); i++) {
                int n = Tableau::count(parts[i]);

                Matrix m (n, n);
                m.zeros();
                coeffs.push_back(m);
        }
}

FourierFunc FourierFunc::zero(int nelements)
{
        if (nelements < 1)
                throw "Not enough elements";

        FourierFunc f(nelements);
        f.parts = Partition::list(nelements);

        f.fill_zero();

        return f;
}

int FourierFunc::num_elements() const
{
        return nelements;
}

/// \todo test right multiplication
FourierFunc FourierFunc::shift(const Permutation &p0, const Permutation &p1) const
{
        FourierFunc ft (*this);
        vector<Partition> parts = Partition::list(nelements);

        /// \todo fix this for band-limiting
        for (unsigned int i = 0; i < ft.coeffs.size(); i++)
                ft.coeffs[i] = mult_irrep_right(parts[i], p1,
                                   mult_irrep_left(parts[i], p0.inverse(),
                                                   ft.coeffs[i]));

        p1.inverse();

        return ft;
}

/// \todo precompute intertwining operators
/// \todo what if not all the dominating partitions are being kept?
Matrix FourierFunc::marginals(Partition part)
{
        vector<Partition> dom_parts;

        for (unsigned int i = 0; i < parts.size(); i++) {
                dom_parts.push_back(parts[i]);

                if (parts[i] == part)
                        break;
        }

        Matrix C = marg_intertwine(part, dom_parts);

        return transpose(C) * assemble_marg_coeffs(part) * C;
}

Matrix FourierFunc::assemble_marg_coeffs(Partition part) const
{
        vector<Matrix> blocks;

        for (unsigned int i = 0; i < parts.size(); i++) {
                if (!parts[i].dominates(part))
                        continue;

                int k = kostka(part, parts[i]);

                blocks.push_back(multi_direct_sum(coeffs[i], k));

                if (parts[i] == part)
                        return direct_sum(blocks);
        }

        // might not include part if band-limited
        return direct_sum(blocks);
}
