/*! \file fourierfunc.h
 *  \brief Declaration of FourierFunc class for band-limited functions on 
 *   S_n stored as Fourier coefficients
 */

#ifndef FOURIERFUNC_H
#define FOURIERFUNC_H

#include "partition.h"
//#include "matrix.h"

#include <itpp/itbase.h>

#define Matrix itpp::mat

class Func;

/// A function on S_n stored as (band-limited) Fourier coefficients
/*! \todo work out how band-limiting will operate
 *  \todo fix Matrix workaround
 */
class FourierFunc {
public:
        /// Create a zero function with coefficients stored at ind_parts
        static FourierFunc zero(std::vector<Partition> ind_parts);

        /// Compute the "naive" inverse Fourier transform
        /*! Thus is a direct translation of the definition, i.e., a sum over
         *  "matrix dot products", and so operates in Theta(n!^2). This exists
         *  for demonstration; in practice inverse_fft should be used instead
         */
        Func inverse_fourier() const;


        /// Compute the inverse Fourier transform using Clausen's FFT
        /*! \todo think about running time of this */
        Func inverse_fft() const;

        /// Shift this function in the Fourier domain
        /*! If this is the Fourier transform of f, the result is the Fourier
         *  transform of f(p) = f(p0^-1 p p1) 
         */
        FourierFunc shift(const Permutation &p0, const Permutation &p1) const;

        /*! \brief Compute the matrix of marignals defined by the tabloids on
         *  partition part
         */
        Matrix marginals(Partition part);

        /// A string representation of this function as coefficient matrices
        std::string str() const;

        /// The matrix of Fourier coefficients at the partition part  
        Matrix at(const Partition &part) const;

        /// Set the ith stored matrix of coefficients to m
        void set(int i, Matrix m);

        /// Set the matrix of coefficients at part to m
        void set(const Partition &part, Matrix m);

        /// Create a full (not band-limited) zero function on S_n
        static FourierFunc zero(int nelements);

        /// Return n where this a function over S_n
        int num_elements() const;

        /// Return the partitions at which Fourier coefficients are stored
        std::vector<Partition> ind_parts();
private:
        /// Construct an unitialized function with no stored coefficents
        FourierFunc(int _nelements);

        Matrix assemble_marg_coeffs(Partition p) const;

        /// Make this function the zero function
        void fill_zero();

        /// This is a function on S_nelements
        int nelements;

        /// The partitions at which coefficients are stored
        std::vector<Partition> parts;

        /// The stored coefficients corresponding to the partitions in parts
        std::vector<Matrix> coeffs;

        friend FourierFunc *uniform_fourier(int n);
};

#undef Matrix

#endif
