/*! \file matrix.cpp
 *  \brief Implementation of Matrix class
 */

#include "matrix.h"

#include <cstring>
#include <sstream>
#include <iomanip>

// Not safe for side effects
#define MIN(x,y) (((x) < (y)) ? (x) : (y))

using namespace std;

Matrix::Matrix()
 : mat(0)
{

}

Matrix::Matrix(const Matrix &m)
{
        *this = m;
}

Matrix &Matrix::operator = (const Matrix &m)
{
        rows = m.rows;
        cols = m.cols;
        mat = new mat_val_t[rows * cols];
        memcpy(mat, m.mat, sizeof(mat_val_t) * rows*cols);

        return *this;
}

Matrix::Matrix(int r, int c)
{
        rows = r;
        cols = c;
        mat = new mat_val_t[rows * cols];
}

Matrix::~Matrix()
{
        delete [] mat;
}

Matrix Matrix::identity(int n)
{
        Matrix m(n, n);
        
        for (int i = 0; i < n; i++)
                for (int j = 0; j < n; j++)
                        if (i == j)
                                m.mat[i*n + j] = 1;
                        else
                                m.mat[i*n + j] = 0;

        return m;
}

Matrix Matrix::zero(int r, int c)
{
        Matrix m(r, c);

        for (int i = 0; i < r; i++)
                for (int j = 0; j < c; j++)
                        m.mat[i*c + j] = 0;

        return m;
}

Matrix Matrix::operator * (mat_val_t s) const
{
        Matrix m(*this);

        for (int i = 0; i < rows; i++)
                for (int j = 0; j < cols; j++)
                        m.mat[i*cols + j] *= s;

        return m;
}

/// \todo check dimensions
Matrix Matrix::operator += (const Matrix &m)
{
        for (int i = 0; i < rows; i++)
                for (int j = 0; j < cols; j++)
                        mat[i*cols + j] += m.mat[i*cols + j];
        return *this;
}

Matrix::mat_val_t Matrix::ele(int i, int j) const
{
        return mat[i * cols + j];
}

Matrix::mat_val_t Matrix::matrix_dot(const Matrix &m) throw (DimensionMismatch)
{
        if (rows != m.rows || cols != m.cols)
                throw DimensionMismatch();

        mat_val_t sum = 0;

        for (int i = 0; i < rows; i++)
                for (int j = 0; j < cols; j++)
                        sum += this->ele(i,j) * m.ele(i,j);

        return sum;
}

void Matrix::set(int r, int c, mat_val_t val)
{
        mat[r*cols + c] = val;
}

Matrix Matrix::submat(int r0, int c0, int r, int c) const
{
        Matrix m(r, c);

        for (int i = 0; i < r; i++)
                for (int j = 0; j < c; j++)
                        m.set(i, j, ele(r0 + i, c0 + j));

        return m;
}

Matrix Matrix::direct_sum(vector<Matrix> ms)
{
        int r = 0;
        int c = 0;
        for (unsigned int i = 0; i < ms.size(); i++) {
                r += ms[i].num_rows();
                c += ms[i].num_cols();
        }
        
        Matrix m = zero(r, c);

        int roff = 0;
        int coff = 0;
        for (unsigned int i = 0; i < ms.size(); i++) {
                for (int j = 0; j < ms[i].num_rows(); j++)
                        for (int k = 0; k < ms[i].num_cols(); k++)
                                m.set(j + roff, k + coff, ms[i].ele(j, k));
                roff += ms[i].num_rows();
                coff += ms[i].num_cols();
        }

        return m;
}

int Matrix::num_rows() const
{
        return rows;
}

int Matrix::num_cols() const
{
        return cols;
}

Matrix::mat_val_t Matrix::trace() const
{
        mat_val_t sum;

        sum = 0;
        for (int i = 0; i < MIN(rows, cols); i++)
                sum += mat[i * cols + i];

        return sum;
}

string Matrix::str() const
{
        ostringstream out;

        for (int i = 0; i < rows; i++) {
                out << "[ ";
                for (int j = 0; j < cols; j++)
                        out << setprecision(4) << setw(6) << ele(i,j) << " ";
                out << "]" << endl;
        }

        return out.str();
}
