/*! \file partition.cpp
 *  \brief Implementation of partitions and related things
 */

#include "partition.h"
#include "util.h"

#include <cstring>
#include <sstream>
#include <iostream>
#include <assert.h>

#define MIN(x, y) ((x) < (y) ? (x) : (y))

using namespace std;

Partition::Partition()
 : n(0), p(0), parts(0)
{

}

Partition::Partition(const Partition &part)
{
        *this = part;
}

/*! \todo have the constructor sort _parts */
Partition::Partition(int _p, int *_parts) throw (InvalidPartition)
{
        if (_p < 0)
                throw InvalidPartition();
        
        p = _p;
        parts = new int[p];
        n = 0;
        for (int i = 0; i < p; i++) {
                parts[i] = _parts[i];
                n += parts[i];
        }
}

Partition &Partition::operator = (const Partition &part)
{
        p = part.p;
        n = part.n;
        parts = new int[p];
        memcpy(parts, part.parts, sizeof(int)*p);

        return *this;
}

Partition::~Partition()
{
       delete [] parts;
}

int Partition::operator () (int i) const
{
        /// \todo error checking
        return parts[i];
}

int Partition::num() const
{
        return n;
}

int Partition::num_parts() const
{
        return p;
}

int Partition::hook_len(int i, int j) const
{
        int k;
        for (k = i; k < p; k++)
                if ((*this)(k) <= j)
                        break;
        return (*this)(i) - j + k - i - 1;
}

/// \todo check that this is not a partition of 1
vector<Partition> Partition::minus() const
{
        vector<Partition> parts_minus;

        for (int i = 0; i < p - 1; i++) {
                if (parts[i] > parts[i + 1])
                        parts_minus.push_back(decr(i));
        }

        parts_minus.push_back(decr(p - 1));

        return parts_minus;
}

bool tabloid_equal(const Partition &part,
                   const Permutation &p0,
                   const Permutation &p1)
{
        int offset = 0;
        for (int i = 0; i < part.num_parts(); i++) {
                int *seq = new_seq(offset, part(i));

                if (!p1.inverse().of(p0).setmapsto(part(i), seq, seq)) {
                        delete [] seq;
                        return false;
                }

                offset += part(i);

                delete [] seq;
        }

        return true;
}

string Partition::str() const
{
        ostringstream out;

        out << "(";
        for (int i = 0; i < p; i++) {
                out << parts[i];
                if (i != p - 1)
                        out << ", ";
        }
        out << ")";

        return out.str();
}

string Partition::repr() const
{
        ostringstream out;

        out << "Partition([";
        for (int i = 0; i < p; i++) {
                out << parts[i];
                if (i != p - 1)
                        out << ", ";
        }
        out << "])";

        return out.str();
}

Partition Partition::decr(int i) const
{
        if (parts[i] == 1) {
                int new_parts[p - 1];
                for (int i = 0; i < p - 1; i++)
                        new_parts[i] = parts[i];
                return Partition(p - 1, new_parts);
        }

        Partition part (*this);
        part.n--;
        part.parts[i]--;
        return part;
}

Partition Partition::subpart(int start, int end) const
{
        return Partition(end - start, parts + start);
}

Partition Partition::concat(const Partition &part) const
{
        int new_parts[p + part.p];

        int i;
        for (i = 0; i < p; i++)
                new_parts[i] = parts[i];
        for (; i < p + part.p; i++)
                new_parts[i] = part.parts[i - p];

        return Partition(p + part.p, new_parts);
}

vector<Partition> Partition::list_up_to(int n, int max, unsigned int size)
{
        if (n == 0)
                return vector<Partition>(1, Partition(0, NULL));

        vector<Partition> res;
        for (int first = MIN(n, max); first >= 1; first--) {
                vector<Partition> subparts = list_up_to(n - first, first, size);

                for (unsigned int i = 0; i < subparts.size(); i++) {
                        int parts[] = {first};
                        res.push_back(Partition(1, parts)
                                      .concat(subparts[i]));
                }

                assert(size == 0 || res.size() <= size);
                if (size != 0 && res.size() == size)
                        break;
        }

        return res;
}

vector<Partition> Partition::list_up_to(int n, unsigned int size)
{
        return list_up_to(n, n, size);
}

vector<Partition> Partition::list(int n)
{
        return list_up_to(n, n, 0);
}

bool Partition::operator == (const Partition &part) const
{
        if (n != part.n || p != part.p)
                return false;

        for (int i = 0; i < p; i++)
                if (parts[i] != part.parts[i])
                        return false;

        return true;
}

bool Partition::operator != (const Partition &part) const
{
        return !(*this == part);
}

bool Partition::dominates(const Partition &part) const
{
        int this_sum = 0;
        int part_sum = 0;

        for (int i = 0; i < MIN(num_parts(), part.num_parts()); i++) {
                this_sum += this->parts[i];
                part_sum += part(i);
                if (part_sum > this_sum)
                        return false;
        }

        return true;
}

static vector<vector<int> > ord_subsets(vector<int> list, unsigned int n)
{
        if (n == list.size())
                return vector<vector<int> >(1, list);

        vector<vector<int> > list_sets;
        
        if (n == 1) {
                for (unsigned int i = 0; i < list.size(); i++)
                        list_sets.push_back(vector<int>(list.end() - 1 - i,
                                                        list.end() - i));
                return list_sets;
        }

        for (unsigned int i = 0; i <= list.size() - n; i++) {
                vector<int> sublist = list;
                sublist.erase(sublist.end() - 1 - i, sublist.end());
                vector<vector<int> > sublist_sets = ord_subsets(sublist, n - 1);
                for (unsigned int j = 0; j < sublist_sets.size(); j++) {
                        sublist_sets[j].push_back(*(list.end() - 1 - i));
                        list_sets.push_back(sublist_sets[j]);
                }
        }
        
        return list_sets;
}

static vector<int> ord_minus(vector<int> a, vector<int> b)
{
        vector<int>::iterator a_it = a.begin();
        for (vector<int>::iterator it = b.begin(); it != b.end(); it++) {
                while (*a_it != *it)
                        a_it++;
                a.erase(a_it);
        }

        return a;
}

static vector<vector<int> > make_tabloid_list(const Partition &part,
                                             vector<int> avail)
{
        const int p = part.num_parts();

        if (p == 1)
                return vector<vector<int> >(1, avail);

        vector<vector<int> > last_list = ord_subsets(avail, part(p - 1));
        vector<vector<int> > tab_list;
        for (unsigned int i = 0; i < last_list.size(); i++) {
                vector<int> remain = ord_minus(avail, last_list[i]);

                vector<vector<int> > sublist =
                        make_tabloid_list(part.subpart(0, p-1), remain);
                for (unsigned int j = 0; j < sublist.size(); j++) {
                        sublist[j].insert(sublist[j].end(),
                                      last_list[i].begin(), last_list[i].end());
                        tab_list.push_back(sublist[j]);
                }
        }
        
        return tab_list;
}

static int *new_list_from_vector(vector<int> v)
{
        int *list = new int[v.size()];

        for (unsigned int i = 0; i < v.size(); i++)
                list[i] = v[i];

        return list;
}

vector<Permutation> tabloid_list(const Partition &part)
{
        vector<Permutation> p_list;
        
        vector<int> vals;

        for (int i = 0; i < part.num(); i++)
                vals.push_back(i);

        vector<vector<int> > val_list = make_tabloid_list(part, vals);

        for (unsigned int i = 0; i < val_list.size(); i++) {
                int *p_vals = new_list_from_vector(val_list[i]);

                p_list.push_back(Permutation(part.num(), p_vals));

                delete [] p_vals;
        }

        return p_list;
}

Tableau::Tableau()
 : part (0, 0), tab (0)
{

}

Tableau::Tableau(const Tableau &t)
 : part (t.part)
{
        tab = new int[part.num()];
        memcpy(tab, t.tab, sizeof(int) * part.num());
}

Tableau::Tableau(const Partition &_part)
 : part (_part)
{
        tab = new int[part.num()];
}

Tableau::~Tableau()
{
        delete [] tab;
}

int Tableau::col(int i) const
{
        int ret = 0;
        for (int j = 0; j < i; j++)
                if (tab[j] == tab[i])
                        ret++;

        return ret;
}

int *Tableau::new_explicit_list() const
{
        int offsets[part.num_parts()];
        int *list = new int[part.num()];

        offsets[0] = 0;
        for (int i = 1; i < part.num_parts(); i++)
                offsets[i] = offsets[i - 1] + part(i - 1);

        for (int i = 0; i < part.num(); i++) {
                list[offsets[tab[i]]] = i;
                offsets[tab[i]]++;
        }

        return list;
}

string Tableau::str() const
{
        ostringstream out;
        int ind;
        int *list = new_explicit_list();

        ind = 0;
        for (int i = 0; i < part.num_parts(); i++) {
                for (int j = 0; j < part(i); j++) {
                        out << list[ind] << " ";
                        ind++;
                }
                if (i != part.num_parts() - 1)
                        out << endl;
        }

        delete [] list;

        return out.str();
}

pair<int, int> Tableau::pos(int i) const
{
        return make_pair(tab[i], col(i));
}

int Tableau::axial_dist(int i, int j) const
{
        pair<int, int> i_pos = pos(i);
        pair<int, int> j_pos = pos(j);

        return j_pos.second - i_pos.second - (j_pos.first - i_pos.first);
}

Tableau Tableau::swap(int i) const
{
        Tableau t(*this);

        t.tab[i] = tab[i + 1];
        t.tab[i + 1] = tab[i];

        return t;
}

bool Tableau::operator == (const Tableau &t)
{
        if (part != t.part)
                return false;
        
        for (int i = 0; i < part.num(); i++)
                if (tab[i] != t.tab[i])
                        return false;

        return true; 
}

int Tableau::count(const Partition &part)
{
        int hook_prod = 1;

        for (int i = 0; i < part.num_parts(); i++)
                for (int j = 0; j < part(i); j++)
                        hook_prod *= part.hook_len(i, j);

        return (int)fact(part.num()) / hook_prod;
}

vector<Tableau> Tableau::list(const Partition &part)
{       
        if (part.num() == 1) {
                Tableau t(part);
                int i;
                for (i = 0; part(i) == 0; i++)
                        ;
                t.tab[0] = i;
                return vector<Tableau>(1, t);
        }

        vector<Tableau> tableaux;

        // look for the next place where the largest number can go
        for (int i = 0; i < part.num_parts(); i++) {
                // skip this row if it doesn't end in a "corner"
                if (i != part.num_parts() - 1 && part(i) == part(i + 1))
                        continue;

                vector<Tableau> subtableaux = Tableau::list(part.decr(i));

                for (unsigned int j = 0; j < subtableaux.size(); j++) {
                        Tableau t(part);
                        t.tab[part.num() - 1] = i;
                        for (int k = 0; k < part.num() - 1; k++)
                                t.tab[k] = subtableaux[j].tab[k];
                        tableaux.push_back(t);
                }
        }

        return tableaux;
}

static vector<Partition> fill_ss_tableau(const Partition &part,
                                         int n, int right)
{
        vector<Partition> subparts;

        for (int i = 0; i < part.num_parts(); i++) {
                // skip this row if it doesn't end in a "corner"
                if (i != part.num_parts() - 1 && part(i) == part(i + 1))
                        continue;
                // skip this row if the end is too far right
                if (part(i) >= right)
                        continue;

                // put the number in the last box in this row
                Partition subpart = part.decr(i);
                // fill in the rest of the numbers
                if (n == 1) {
                        subparts.push_back(subpart);
                        continue;
                }

                vector<Partition> subparts_i =
                        fill_ss_tableau(subpart, n - 1, part(i));
                for (unsigned int j = 0; j < subparts_i.size(); j++)
                        subparts.push_back(subparts_i[j]);
        }
        return subparts;
}

int kostka(const Partition &m_part, const Partition &i_part)
{
        /// \todo should we check here that the Kostka number is nonzero?
        if (m_part.num_parts() == 1) {
                if (i_part.num_parts() == 1)
                        return 1;
                else
                        return 0;
        }

        // determine all the ways of placing the largest numbers in the
        // semistandard tableau of shape m_part and content i_part, then
        // recurse
        int k = 0;

        vector<Partition> subparts =
               fill_ss_tableau(i_part, m_part(m_part.num_parts() - 1),
                               i_part(0) + 1);
        for (unsigned int i = 0; i < subparts.size(); i++)
                k += kostka(m_part.subpart(0, m_part.num_parts() - 1), subparts[i]);

        return k;
}
