from props import *
from numpy import abs, sum, random, hstack, ones
import unittest

def sampleDistribution():
    f = Func.zero(3)
    f.set(Permutation.identity(3), 1./3)
    f.set(Permutation([1, 0, 2]), 1./6)
    f.set(Permutation([0, 2, 1]), 1./3)
    f.set(Permutation([1, 2, 0]), 1./6)
    return f

def randomDistribution(n):
    f = Func.zero(n)
    r = random.random(fact(n) - 1)
    r.sort()
    dvals = hstack((r, [1])) - hstack(([0], r))

    for i in range(fact(n)):
        p = Permutation.from_index(n, i)
        f.set(p, dvals[i])

    return f

class TestFourierDomainModels(unittest.TestCase):
    def testUniform(self):
        fc = uniform_fourier(5)
        m = fc.marginals(Partition([4, 1]))
        self.assertAlmostEqual(sum(abs(m - 1/5. * ones((5, 5)))), 0)

    def testZerosOutsideBand(self):
        fc = uniform_fourier(4)
        for part in Partition.list(4)[1:]:
            self.assertAlmostEqual(sum(abs(fc.at(part))), 0)

class TestFourierTransform(unittest.TestCase):

    def testFFTConsistency(self):
        f = sampleDistribution()
        g = f.fft().inverse_fft()
        n = f.num_elements()
        for p in map(lambda i: Permutation.from_index(n, i), range(fact(n))):
            self.assertAlmostEqual(f(p), g(p))

    def testFFTMarginals(self):
        fs = [sampleDistribution()] + map(randomDistribution, range(1, 5))
        for f in fs:
            fc = f.fft()
            for part in Partition.list(f.num_elements()):
                marg = f.marginals(part)
                marg_ft = fc.marginals(part)
                self.assertAlmostEqual(sum(abs(marg - marg_ft)), 0)
        
class TestPartitions(unittest.TestCase):

    def testExplicitConstruction(self):
        # note: numpy's random.randint, not Python's
        parts = map(lambda i: random.randint(1, 10), range(random.randint(1, 6)))
        part = Partition(parts)
        parts.sort()
        for i in range(len(parts)):
            self.assertEqual(parts[i], part(i))
        self.assertEqual(part.num(), sum(parts))
        self.assertEqual(part.num_parts, parts.len())

    def testBadConstruction(self):
        self.assertRaises(InvalidPartition, Partition, [6, 2, 0])
        self.assertRaises(InvalidPartition, Partition, [7, -1])

class TestPermutations(unittest.TestCase):

    def testExplicitConstruction(self):
        map = [1, 0, 2, 3, 5, 4]
        p = Permutation(map)
        for i in range(len(map)):
            self.assertEqual(map[i], p(i))

    def testIdentityConstruction(self):
        p = Permutation.identity(5)
        for i in range(5):
            self.assertEqual(i, p(i))

    def testComposition(self):
        map0 = [0, 1, 3, 2]
        map1 = [1, 3, 2, 0]
        mapcomp = [1, 2, 3, 0]

        p = Permutation(map0).of(Permutation(map1))

        for i in range(4):
            self.assertEqual(p(i), mapcomp[i])

    def testInverse(self):
        map = [3, 0, 2, 1]
        mapinv = [1, 3, 2, 0]

        p = Permutation(map).inverse()

        for i in range(4):
            self.assertEqual(p(i), mapinv[i])

    def testDirectSum(self):
        map0 = [0, 2, 1]
        map1 = [4, 0, 3, 1, 2]
        mapsum = [0, 2, 1, 7, 3, 6, 4, 5]

        p = Permutation(map0).direct_sum(Permutation(map1))

        for i in range(8):
            self.assertEqual(p(i), mapsum[i])

    def testListAgreesWithFromIndex(self):
        n = 5
        for (p, j) in zip(Permutation.list(n), range(fact(n))):
            p1 = Permutation.fromIndex(j)
            for i in range(n):
                self.assertEqual(p(i), p1(i))

    def testBadConstruction(self):
        self.assertRaises(InvalidPermutation, Permutation, [0, 0, 1])
        self.assertRaises(InvalidPermutation, Permutation, [-1, 0, 1])
        self.assertRaises(InvalidPermutation, Permutation, [0, 2])
        self.assertRaises(InvalidPermutation, Permutation, [])

    def testBadIndex(self):
        p = Permutation.from_index(5, 8)
        self.assertRaises(IndexOutOfBounds, p, -1)
        self.assertRaises(IndexOutOfBounds, p, 5)

    def testBadComposition(self):
        p0 = Permutation.from_index(5, 7)
        p1 = Permutation.from_index(6, 9)
        self.assertRaises(PermutationMismatch, p0.of, p1)


if __name__ == "__main__":
    unittest.main()
